from metrics_layer.core.exceptions import AccessDeniedOrDoesNotExistException, QueryError
from .base import MetricsLayerBase
from .join import Join
from .set import Set


class Explore(MetricsLayerBase):
    def __init__(self, definition: dict = {}, project=None) -> None:
        if definition.get("from") is not None:
            definition["from_"] = definition["from"]
        elif definition.get("view_name") is not None:
            definition["from_"] = definition["view_name"]
        else:
            definition["from_"] = definition["name"]

        self.project = project
        self._view_names = []
        self.validate(definition)
        super().__init__(definition)

    @property
    def week_start_day(self):
        if self.model.week_start_day:
            return self.model.week_start_day.lower()
        return "monday"

    def validate(self, definition: dict):
        required_keys = ["name", "model", "from_"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Explore missing required key {k}")

    def printable_attributes(self):
        to_print = ["name", "type", "from", "label", "group_label", "model_name", "join_names"]
        attributes = self.to_dict()
        attributes["type"] = "explore"
        attributes["from"] = attributes["from_"]
        attributes["join_names"] = [j["name"] for j in attributes.get("joins", [])]
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    def validate_fields(self):
        errors = []
        if not self.valid_name(self.name):
            errors.append(self.name_error("explore", self.name))

        for join in self.joins():
            errors.extend(join.collect_errors())

        for view_name in self.view_names():
            try:
                view = self.project.get_view(view_name, explore=self)
            except Exception:
                errors.append(f"View {view_name} cannot be found in explore {self.name}")
                continue
            try:
                view.sql_table_name
            except QueryError as e:
                errors.append(f"{str(e)} in explore {self.name}")

            referenced_fields = view.referenced_fields()
            view_errors = view.collect_errors()
            errors.extend(
                [
                    f"Could not locate reference {field} in view {view_name} in explore {self.name}"
                    for field in referenced_fields
                    if isinstance(field, str)
                ]
            )
            errors.extend(view_errors)

        return list(sorted(list(set(errors))))

    def _all_joins(self):
        joins = []
        for j in self._definition.get("joins", []):
            join = Join(j, explore=self, project=self.project)
            if self.project.can_access_join(join, explore=self):
                joins.append(join)
        return joins

    def view_names(self):
        if self._view_names:
            return self._view_names
        self._view_names = [self.from_] + [j.from_ for j in self.joins()]
        return self._view_names

    def join_names(self):
        return [self.name] + [j.name for j in self._all_joins()]

    def field_names(self):
        # This function is for the explore `fields` parameter, to resolve all the sets into field names
        if self.fields is None:
            return self.fields

        set_definition = {"name": "NA", "fields": self.fields, "explore_name": self.name}
        explore_set = Set(set_definition, project=self.project, explore=self)
        return explore_set.field_names()

    def get_join(self, join_name: str, by_view_name: bool = False):
        if by_view_name:
            return next((j for j in self.joins() if j.from_ == join_name), None)
        return next((j for j in self.joins() if j.name == join_name), None)

    def joins(self):
        output = []
        for join in self._all_joins():
            if join.is_valid():
                output.append(join)
        return output

    def explore_fields(self, show_hidden: bool, expand_dimension_groups: bool, show_excluded: bool):
        try:
            view = self.project.get_view(self.from_, explore=self)
            fields = view.fields(show_hidden, expand_dimension_groups)
        except AccessDeniedOrDoesNotExistException:
            fields = []

        for join in self.joins():
            fields.extend(join.join_fields(show_hidden, expand_dimension_groups, show_excluded))

        explore_field_names = self.field_names()
        if explore_field_names and not show_excluded:
            return [f for f in fields if f.id() in explore_field_names]
        return fields



import sqlparse

from metrics_layer.core.convert import MQLConverter
from metrics_layer.core.parse import ProjectLoader
from metrics_layer.core.parse.connections import BaseConnection
from metrics_layer.core.sql import QueryRunner, SQLQueryResolver
from metrics_layer.core.sql.query_errors import ParseError
from metrics_layer.core.exceptions import QueryError


class DBConnectionError(Exception):
    pass


class MetricsLayerConnection:
    def __init__(
        self,
        location: str = None,
        branch: str = "master",
        project=None,
        connections: list = [],
        user: dict = None,
        **kwargs,
    ):
        self.location, self.branch, self._raw_connections = location, branch, connections
        self._raw_project = project
        self.kwargs = kwargs
        self._user = user
        self.branch_options = None
        self._project = None

    def set_user(self, user: dict):
        self._user = user
        self.project.set_user(self._user)

    def load(self):
        if self.location is not None:
            self._loader = ProjectLoader(self.location, self.branch, self._raw_connections)
            self._project = self._loader.load()
            self._project.set_user(self._user)
            self.branch_options = self._loader.get_branch_options()
        elif self._raw_project is not None:
            self._project = self._raw_project
            self._project.set_user(self._user)
            self.branch_options = []
        else:
            raise DBConnectionError(
                "No project or location specified. Pass either the location "
                "(a path or a github url) or a project object."
            )

    @property
    def profiles_path(self):
        return ProjectLoader.profiles_path()

    @property
    def project(self):
        if self._project is None:
            self.load()
        return self._project

    def get_branch_options(self):
        if self.branch_options is not None:
            return self.branch_options
        self.load()
        return self.branch_options

    @property
    def connections(self):
        self._connections = ProjectLoader.load_connections(self._raw_connections)
        return self._connections

    def list_connections(self, names_only=False):
        connections = self.connections
        if names_only:
            return [c.name for c in connections]
        return connections

    def get_connection(self, connection_name: str):
        return next((c for c in self.list_connections() if c.name == connection_name), None)

    def query(
        self,
        metrics: list = [],
        dimensions: list = [],
        funnel: dict = {},
        where: list = [],
        having: list = [],
        order_by: list = [],
        sql: str = None,
        **kwargs,
    ):
        query, connection = self.get_sql_query(
            sql=sql,
            metrics=metrics,
            dimensions=dimensions,
            funnel=funnel,
            where=where,
            having=having,
            order_by=order_by,
            **{**self.kwargs, **kwargs},
            return_connection=True,
        )
        df = self.run_query(query, connection, **kwargs)
        return df

    def get_sql_query(
        self,
        metrics: list = [],
        dimensions: list = [],
        funnel: dict = {},
        where: list = [],
        having: list = [],
        order_by: list = [],
        sql: str = None,
        **kwargs,
    ):
        model = self._get_model_for_query(kwargs.get("model_name"), metrics, dimensions)
        connection = self.get_connection(model.connection)
        query_type = self._get_query_type(connection, kwargs)
        connection_schema = self._get_connection_schema(connection)
        self.project.set_connection_schema(connection_schema)
        if sql:
            converter = MQLConverter(
                sql,
                project=self.project,
                model=model,
                query_type=query_type,
                **{**self.kwargs, **kwargs},
            )
            query = converter.get_query()
        else:
            resolver = SQLQueryResolver(
                metrics=metrics,
                dimensions=dimensions,
                funnel=funnel,
                where=where,
                having=having,
                order_by=order_by,
                project=self.project,
                model=model,
                query_type=query_type,
                **{**self.kwargs, **kwargs},
            )
            query = resolver.get_query()

        if kwargs.get("pretty", False):
            query = self.pretty_sql(query)

        if kwargs.get("return_connection", False):
            return query, connection

        if kwargs.get("return_query_kind", False):
            return query, resolver.query_kind
        return query

    def run_query(self, query: str, connection: BaseConnection, **kwargs):
        runner = QueryRunner(query, connection)
        df = runner.run_query(**{**self.kwargs, **kwargs})
        return df

    def define(self, metric: str, **kwargs):
        field = self.project.get_field(metric)
        connection = self.get_connection(field.view.model.connection)
        query_type = self._get_query_type(connection, kwargs)
        return field.sql_query(query_type)

    def list_fields(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        if names_only:
            return [m.name for m in all_fields]
        return all_fields

    def get_field(self, field_name: str, view_name: str = None):
        models = self.list_models()
        kws = {}
        if len(models) == 1:
            kws = {"model": models[0]}
        return self.project.get_field(field_name, view_name=view_name, **kws)

    def list_metrics(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        metrics = [f for f in all_fields if f.field_type == "measure"]
        if names_only:
            return [m.name for m in metrics]
        return metrics

    def get_metric(self, metric_name: str, view_name: str = None):
        metrics = self.list_metrics(view_name=view_name)
        try:
            metric = next((m for m in metrics if m.equal(metric_name)))
            return metric
        except ParseError as e:
            raise e(f"Could not find metric {metric_name} in the project config")

    def list_dimensions(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        dimensions = [f for f in all_fields if f.field_type in {"dimension", "dimension_group"}]
        if names_only:
            return [d.name for d in dimensions]
        return dimensions

    def get_dimension(self, dimension_name: str, view_name: str = None):
        dimensions = self.list_dimensions(view_name=view_name)
        try:
            dimension = next((d for d in dimensions if d.equal(dimension_name)))
            return dimension
        except ParseError as e:
            raise e(f"Could not find dimension {dimension_name} in the project config")

    def list_views(self, names_only=False):
        views = self.project.views()
        if names_only:
            return [v.name for v in views]
        return views

    def get_view(self, view_name: str):
        return self.project.get_view(view_name)

    def list_models(self, names_only=False):
        models = self.project.models()
        if names_only:
            return [m.name for m in models]
        return models

    def get_model(self, model_name: str):
        return self.project.get_model(model_name)

    def list_dashboards(self, names_only=False):
        dashboards = self.project.dashboards()
        if names_only:
            return [d.name for d in dashboards]
        return dashboards

    def get_dashboard(self, dashboard_name: str):
        return self.project.get_dashboard(dashboard_name)

    def get_all_profiles(self, names_only: bool = False):
        raise NotImplementedError()

    def _get_model_for_query(self, model_name: str = None, metrics: list = [], dimensions: list = []):
        models = self.project.models()

        # If you specify the model that's top priority
        if model_name:
            return self.project.get_model(model_name)
        # Otherwise, if there's only one option, we use that
        elif len(models) == 1:
            return models[0]
        # Finally, check views for models
        else:
            return self._derive_model(metrics, dimensions)

    def _derive_model(self, metrics: list, dimensions: list):
        all_model_names = {f.view.model_name for f in metrics + dimensions}

        if len(all_model_names) == 0:
            raise QueryError(
                "No models found in this data model. Please specify a model "
                "to connect a data warehouse to your data model."
            )
        elif len(all_model_names) == 1:
            return self.project.get_model(list(all_model_names)[0])
        else:
            raise QueryError(
                "More than one model found in this data model. Please specify a model "
                "to use by either passing the name of the model using 'model_name' parameter or by  "
                "setting the `model_name` property on the view."
            )

    def _get_query_type(self, connection, kwargs: dict):
        if "query_type" in kwargs:
            return kwargs.pop("query_type")
        elif connection:
            return connection.type
        else:
            raise QueryError(
                "Could not determine query_type. Please have connection information for "
                "your warehouse in the configuration or explicitly pass the "
                "'query_type' argument to this function"
            )

    def _get_connection_schema(self, connection):
        if connection is not None:
            return getattr(connection, "schema", None)
        return None

    @staticmethod
    def get_connections_from_profile(profile_name: str):
        return ProjectLoader.get_connections_from_profile(profile_name)

    @staticmethod
    def pretty_sql(sql: str, keyword_case="lower"):
        return sqlparse.format(sql, reindent=True, keyword_case=keyword_case)



try:
    import snowflake.connector
except ModuleNotFoundError:
    pass

try:
    from google.cloud import bigquery
    from google.oauth2 import service_account
except ModuleNotFoundError:
    pass

try:
    import redshift_connector
except ModuleNotFoundError:
    pass

from metrics_layer.core.parse.connections import (
    BaseConnection,
    ConnectionType,
    MetricsLayerConnectionError,
)


class QueryRunner:
    def __init__(self, query: str, connection: BaseConnection):
        self.query = query
        self.connection = connection
        self._query_runner_lookup = {
            ConnectionType.snowflake: self._run_snowflake_query,
            ConnectionType.bigquery: self._run_bigquery_query,
            ConnectionType.redshift: self._run_redshift_query,
        }
        if self.connection.type not in self._query_runner_lookup:
            supported = list(self._query_runner_lookup.keys())
            raise MetricsLayerConnectionError(
                f"Connection type {self.connection.type} not supported, supported types are {supported}"
            )

    # 3 min timeout default set in seconds (aborts query after timeout)
    def run_query(self, timeout: int = 180, **kwargs):
        query_runner = self._query_runner_lookup[self.connection.type]
        df = query_runner(
            timeout=timeout,
            raw_cursor=kwargs.get("raw_cursor", False),
            run_pre_queries=kwargs.get("run_pre_queries", True),
            start_warehouse=kwargs.get("start_warehouse", True),
        )
        return df

    def _run_snowflake_query(
        self, timeout: int, raw_cursor: bool, run_pre_queries: bool, start_warehouse: bool
    ):
        snowflake_connection = self._get_snowflake_connection(self.connection)
        if run_pre_queries:
            self._run_snowflake_pre_queries(snowflake_connection)
        elif start_warehouse:
            self._run_snowflake_pre_queries(snowflake_connection, warehouse_only=True)
        cursor = snowflake_connection.cursor()
        cursor.execute(self.query, timeout=timeout)

        snowflake_connection.close()
        if raw_cursor:
            return cursor
        df = cursor.fetch_pandas_all()
        return df

    def _run_redshift_query(
        self, timeout: int, raw_cursor: bool, run_pre_queries: bool, start_warehouse: bool
    ):
        redshift_connection = self._get_redshift_connection(self.connection, timeout=timeout)
        cursor = redshift_connection.cursor()
        cursor.execute(self.query)

        redshift_connection.close()
        if raw_cursor:
            return cursor
        df = cursor.fetch_dataframe()
        return df

    def _run_bigquery_query(
        self, timeout: int, raw_cursor: bool, run_pre_queries: bool, start_warehouse: bool
    ):
        bigquery_connection = self._get_bigquery_connection(self.connection)
        result = bigquery_connection.query(self.query, timeout=timeout, job_retry=None)
        bigquery_connection.close()
        if raw_cursor:
            return result
        df = result.to_dataframe()
        return df

    def _run_snowflake_pre_queries(self, snowflake_connection, warehouse_only: bool = False):
        to_execute = ""
        if self.connection.warehouse:
            to_execute += f"USE WAREHOUSE {self.connection.warehouse};"
        if self.connection.database and not warehouse_only:
            to_execute += f'USE DATABASE "{self.connection.database.upper()}";'
        if self.connection.schema and not warehouse_only:
            to_execute += f'USE SCHEMA "{self.connection.schema.upper()}";'

        if to_execute != "":
            snowflake_connection.execute_string(to_execute)

    @staticmethod
    def _get_snowflake_connection(connection: BaseConnection):
        try:
            return snowflake.connector.connect(
                account=connection.account,
                user=connection.username,
                password=connection.password,
                role=connection.role,
            )
        except (ModuleNotFoundError, NameError):
            raise ModuleNotFoundError(
                "MetricsLayer could not find the Snowflake modules it needs to run the query. "
                "Make sure that you have those modules installed or reinstall MetricsLayer with "
                "the [snowflake] option e.g. pip install metrics-layer[snowflake]"
            )

    @staticmethod
    def _get_redshift_connection(connection: BaseConnection, timeout: int = None):
        try:
            return redshift_connector.connect(
                host=connection.host,
                port=connection.port,
                database=connection.database,
                user=connection.username,
                password=connection.password,
                timeout=timeout,
            )
        except (ModuleNotFoundError, NameError):
            raise ModuleNotFoundError(
                "MetricsLayer could not find the Redshift modules it needs to run the query. "
                "Make sure that you have those modules installed or reinstall MetricsLayer with "
                "the [redshift] option e.g. pip install metrics-layer[redshift]"
            )

    @staticmethod
    def _get_bigquery_connection(connection: BaseConnection):
        try:
            service_account_creds = service_account.Credentials.from_service_account_info(
                connection.credentials
            )
            connection = bigquery.Client(project=connection.project_id, credentials=service_account_creds)
        except (ModuleNotFoundError, NameError):
            raise ModuleNotFoundError(
                "MetricsLayer could not find the BigQuery modules it needs to run the query. "
                "Make sure that you have those modules installed or reinstall MetricsLayer with "
                "the [bigquery] option e.g. pip install metrics-layer[bigquery]"
            )
        return connection



import os
import shutil
from glob import glob
import yaml

import git

from metrics_layer.core import utils

BASE_PATH = os.path.dirname(__file__)


class BaseRepo:
    def get_repo_type(self):
        if self.repo_type:
            return self.repo_type

        project_files = list(self.search(pattern="zenlytic_project.yaml", folders=[self.folder]))
        project_files += list(self.search(pattern="zenlytic_project.yml", folders=[self.folder]))
        if len(project_files) == 1:
            return self.read_yaml_file(project_files[0]).get("mode", "metrics_layer")
        return "metrics_layer"

    def delete(self):
        raise NotImplementedError()

    def search(self, pattern: str, folders: list = []):
        """Example arg: pattern='*.yml'"""
        return [fn for f in folders for fn in self.glob_search(f, pattern) if "venv" not in fn]

    def fetch(self):
        raise NotImplementedError()

    @staticmethod
    def read_yaml_file(path: str):
        with open(path, "r") as f:
            yaml_dict = yaml.safe_load(f)
        return yaml_dict

    @staticmethod
    def glob_search(folder: str, pattern: str):
        return glob(f"{folder}**/{pattern}", recursive=True)


class LocalRepo(BaseRepo):
    def __init__(self, repo_path: str, repo_type: str = None, dbt_path: str = None) -> None:
        self.repo_path = repo_path
        self.dbt_path = dbt_path
        if isinstance(dbt_path, str) and dbt_path[-1] != "/":
            self.dbt_path += "/"
        self.repo_type = repo_type
        self.folder = f"{os.path.join(os.getcwd(), self.repo_path)}/"
        self.branch_options = []

    def fetch(self):
        pass

    def delete(self):
        pass


class GithubRepo(BaseRepo):
    def __init__(self, repo_url: str, branch: str, repo_type: str = None) -> None:
        self.repo_url = repo_url
        self.repo_type = repo_type
        self.repo_name = utils.generate_uuid()
        self.repo_destination = os.path.join(BASE_PATH, self.repo_name)
        self.folder = f"{self.repo_destination}/"
        self.dbt_path = self.folder
        self.branch = branch
        self.branch_options = []

    def fetch(self):
        self.fetch_github_repo(self.repo_url, self.repo_destination, self.branch)

        try:
            dynamic_branch_options = []
            g = git.cmd.Git()
            raw = g.ls_remote(self.repo_url)
            for raw_branch_ref in raw.split("\n"):
                if "/heads/" in raw_branch_ref:
                    clean_branch_ref = raw_branch_ref.split("/heads/")[-1]
                    dynamic_branch_options.append(clean_branch_ref)
            self.branch_options = dynamic_branch_options
        except Exception as e:
            print(f"Exception getting branch options: {e}")
            self.branch_options = [self.branch]

    def delete(self, folder: str = None):
        if folder is None:
            folder = self.folder

        if os.path.exists(folder) and os.path.isdir(folder):
            shutil.rmtree(folder)

    @staticmethod
    def fetch_github_repo(repo_url: str, repo_destination: str, branch: str):
        if os.path.exists(repo_destination) and os.path.isdir(repo_destination):
            shutil.rmtree(repo_destination)
        repo = git.Repo.clone_from(repo_url, to_path=repo_destination, branch=branch, depth=1)
        return repo
